# IKVM Tutorial

- [Setup your Environment](#setup-your-environment)
- [Run a Java Application Dynamically](#run-a-java-application-dynamically)
- [Convert a Java Application to .NET](#convert-a-java-application-to-net)
- [Develop a .NET Application in Java](#develop-a-net-application-in-java)
  - [Step 1: Generate Java stubs](#step-1-generate-java-stubs)
  - [Step 2: Compile the Java source code](#step-2-compile-the-java-source-code)
  - [Step 3: Generate a .NET executable](#step-3-generate-a-net-executable)

## Setup your Environment

This tutorial includes information for both the Windows and Linux platforms. It assumes that Windows users will be using the .NET SDK, and Linux users will be using the Mono SDK.

This tutorial references files in the samples distribution available on the [Releases](https://github.com/ikvm-revived/ikvm/releases) page. Before you begin, prepare your environment by adding the following to your PATH environment variable:

- The directory containing the IKVM executables
- The directory containing the C# compiler (Windows: csc / Mono: mcs). On Windows, this is typically C:\WINDOWS\Microsoft.NET\Framework\v1.1.4322.
- The directory containing a Java compiler (javac or jikes)


## Run a Java Application Dynamically

IKVM includes a Java Virtual Machine implemented in C#. To try it out, navigate to `IKVMROOT\samples\hello` and compile the sample application: 

```console
javac Hello jar cfm hello.jar manifest.mf Hello.class
```

Now, to run the application using the IKVM Virtual Machine, enter the following:

```console
ikvm Hello
```

This command starts the virtual machine, which searches for a file named `Hello.class`. When it finds it, it loads it and executes the bytecode dynamically. You should be prompted to enter your name, and see a brief greeting.

If you experience problems, check the following:

- Check your capitalization: ikvm, like java, requires you to capitalize class names correctly.
- If ikvm reports a [ClassNotFoundException](class-not-found-exception.md), check whether the `CLASSPATH `environment variable is set. If so, try clearing the `CLASSPATH` or adding the current directory to it so ikvm can find the class in the current directory.

You can also execute Java applications in a jar file. Try it out:

```console
ikvm -jar hello.jar
```

See the [ikvm](tools/ikvm.md) reference for more information about `ikvm` command line options.

## Convert a Java Application to .NET

IKVM includes [ikvmc](tools/ikvmc.md), a utility that converts Java `.jar` files to .NET .dll libraries and .exe applications. In this section, you'll convert a Java application to a .NET .exe.

Navigate to `IKVMROOT\samples\hello` and enter the following:

```console
ikvmc hello.jar
```

After the command completes, you should find a `hello.exe` file in the current directory. To execute it:

### Windows / .NET Framework:
- Try running `hello.exe`. If you get a `FileNotFoundException` when the .NET runtime attempts to load the referenced `IKVM.Java.dll`, remember that the .NET Framework expects to find referenced dll's in the application directory or in the Global Assembly Cache. Either install the dll's in the Global Assembly Cache, or copy them to the application directory.

### Linux / Mono:

Run it using the following command:

```console
mono hello.exe
```

## Develop a .NET Application in Java

In this section, you will learn the steps needed to develop .NET applications in Java. Take a look at the source code of `ShowDir.java` -- this is a Java application that uses the .NET API to display a list of files in the current directory.

```java
import cli.System.IO.*;

public class ShowDir{

   public static void main(String[] args){
       String[] files = Directory.GetFiles(".");
       for(String file : files){
           System.out.println(file);
       }
   }
}
```

Notice the imports at the top -- the package names begin with cli.*. These are not packages in the Java API; rather, they are "pseudo" packages that map to .NET namespaces.

### Step 1: Generate Java stubs

IKVM does not come with a Java compiler, so we will compile ShowDir using a standard Java compiler. Since Java compilers can only compile applications that use Java API's, not .NET API's, we have to fool the Java compiler into believing that there is really a Java package named `cli.System.IO`. The [ikvmstub](tools/ikvmstub.md) application helps us do this. It generates Java `.jar` files from .NET dll's. The `.jar` files generated by [ikvmstub](tools/ikvmstub.md) contain Java classes and interfaces that correspond to .NET classes, but don't contain any real code. They contain just enough to satisfy the Java compiler, and allow it to type check the Java application.

Type the following:

```console
ikvmstub mscorlib.dll
```

Note: On a Linux Mono installation, you will have to type the full pathname to `mscorlib.dll`, like this: 

```console
ikvmstub /usr/lib/mscorlib.dll
```

After the command completes, you should find a file named `mscorlib.jar` in the current directory.

### Step 2: Compile the Java source code

Now, we'll compile the Java source code. If you're using `javac`, type the following:

```console
javac -classpath mscorlib.jar ShowDir.java
```

(Substitute jikes for javac if you're using that tool.) Of course you can also use any Java IDE like Eclipse.

After the command completes, you should find `ShowDir.class` in the current directory.

### Step 3: Generate a .NET executable

Now, we'll convert the Java class file to a .NET application. Type the following:

```console
ikvmc -r:mscorlib.dll ShowDir.class
```

After the command completes, you should find `ShowDir.exe` in the current directory. You should be able to execute it successfully. (On Windows .NET, remember to copy the IKVM dll's to the current directory.)

See for [ikvmc](tools/ikvmc.md) and [ClassLoader](class-loader.md) for more details.

## Related

- [ClassLoader](class-loader.md)
- [ClassNotFoundException](class-not-found-exception.md)
- [Download](https://github.com/ikvm-revived/ikvm/releases)
- [ikvm](tools/ikvm.md)
- [ikvmc](tools/ikvmc.md)
- [ikvmstub](tools/ikvmstub.md)
